// Copyright 2022 ZORALab Enterprise (hello@zoralab.com)
// Copyright 2022 "Holloway" Chew, Kean Ho (hollowaykeanho@gmail.com)
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package thelper

import (
	"fmt"
	"math/big"
	"strings"
	"testing"
	"time"
)

type FlushConfig uint

// Flush configurations are the control values for controlling `Flush(...) intensity.
const (
	// AllTypes cleans both `Log` and `Failed` containers.
	AllType FlushConfig = iota + 1

	// LogType only cleans `Log` containers.
	LogType

	// LogType only cleans `Failed` containers.
	FailedType
)

// TController is the interface for THelper to report out all the messages.
//
// It takes 2 basic inputs:
//   1. Errorf - printing by failed cases
//   2. Logf   - printing by passed cases
type TController interface {
	Errorf(format string, args ...interface{})
	Logf(format string, args ...interface{})
}

// THelper is a test helper structure expanding existing `testing` tool.
//
// It offers various assertions, generator and some calculator functions for
// simplifying the testing needs.
//
// This structure has 3 public elements:
//   1. Controller   - holding TController interface like * testing.T
//   2. FailKeyword  - the "FAIL" announcement that allows you to quickly
//                     search the failing cases one by one. It can be unique
//                     to you.
//   3. QuietMode    - instructs the helper to only print failed cases,
//                     leaving passed cases quiet.
//
// This structure has private elements so use NewTHelper(...) function to
// create one instead of using the conventional structre{} method.
type THelper struct {
	Controller  TController
	FailKeyword string
	log         []string
	failed      []string
	QuietMode   bool
}

// NewTHelper creates a properly defined THelper object.
//
// It takes 1 input:
//   1. `t`    - your *testing.T test controller object from standard package.
//
// By default, the public interfacing elements are configured in the following
// manners. They are customizable after the THelper object is created and
// initialized.
func NewTHelper(t *testing.T) *THelper {
	return &THelper{
		Controller:  t,
		FailKeyword: "TESTFAIL",
		log:         []string{},
		failed:      []string{},
	}
}

// Errorf prints an error message reporting function similar to `t.Errorf`.
//
// The messages are queued and only be printed out when `Thelper.Conclude()`
// is called.
func (me *THelper) Errorf(format string, a ...interface{}) {
	var s string

	if me.FailKeyword != "" {
		newFormat := "%v: " + format

		la := append([]interface{}{me.FailKeyword}, a...)
		s = fmt.Sprintf(newFormat, la...)
	} else {
		s = fmt.Sprintf(format, a...)
	}

	me.failed = append(me.failed, s)
}

// Logf prints a log message similar to `t.Logf` function.
//
// The messages are queued and only be printed out when `Thelper.Conclude()`
// is called.
func (me *THelper) Logf(format string, a ...interface{}) {
	s := fmt.Sprintf(format, a...)
	me.log = append(me.log, s)
}

// LogScenario logs a given Scenario as a standard test case reporting.
//
// It utilizes `Thelper.Logf(...)` under the hood to processing the output
// data.
//
// To accomandate flexibility, this function facilitates an optional data map
// parameter that allows tester to log any data alongside the logging process.
// The `string` label of the list is the field name while the value can be
// anything. Note that the value access is **not** thread-safe so it's upto the
// test script to pre-process them before inputting.
//
// The messages are queued and only be printed out when `Thelper.Conclude()`
// is called.
func (me *THelper) LogScenario(s Scenario, data map[string]interface{}) {
	me.Logf("CASE\n%v\n", s.UID)
	me.Logf("TEST TYPE\n%v\n", s.TestType)
	me.Logf("DESCRIPTION\n%v\n", strings.TrimSpace(s.Description))

	// process switches
	me.Logf("SWITCHES")

	for k, v := range s.Switches {
		me.Logf("%-20v = %v", k, v)
	}

	me.Logf("\n")

	// process data
	if len(data) == 0 {
		return
	}

	me.Logf("GOT")

	for k, v := range data {
		me.Logf("%-20v = %#v", k, v)
	}

	me.Logf("\n")
}

// Flush resets all message containers in a THelper structure for next run.
//
// Flush is made available in case of manual intervention needs.
func (me *THelper) Flush(t FlushConfig) {
	switch t {
	case LogType:
		me.log = []string{}
	case FailedType:
		me.failed = []string{}
	case AllType:
		me.log = []string{}
		me.failed = []string{}
	}
}

// Conclude signals the test has ended and pretty print all the stored logs.
//
// Conclude uses the internal TController interface to perform the printing.
// The outcome can be either:
//   1. using `t.Errorf(...)` for logged error/failed messages;
//   2. using `t.Logf(...)` for logged messages.
//
// If QuietMode is true for the outcome #2, `Conclude()` shall not print
// anything.
//
// Upon successful output, the THelper structure shall automatically call
// `Flush(AllType)` function, making it ready for the next test run.
func (me *THelper) Conclude() {
	var log, err string

	if me.Controller == nil {
		goto conclude
	}

	// process error messages
	err = strings.Join(me.failed, "\n")

	// process log messages
	log = strings.Join(me.log, "\n")

conclude:
	switch {
	case me.QuietMode && err == "":
	case me.QuietMode && err != "":
		fallthrough
	case log != "" && err != "":
		me.Controller.Errorf("\n%v\n%v\n\n\n", err, log)
	case log != "":
		me.Controller.Logf("\n%v\n\n\n", log)
	case err != "":
		me.Controller.Errorf("\n%v\n\n\n", err)
	}

	me.Flush(AllType)
}

// ExpectExists asserts a given object existence.
//
// It takes 3 inputs:
//   1. label       - for error logging purposes: asserted object's name.
//   2. subject     - the test subject being asserted.
//   3. expected    - the expectation for the subject existence. `true` means it
//                    is expected to exist.
//
// This function automatically logs the error message into if the test subject
// contradict the expectation.
//
// It returns:
//   1. 0     - passed assertion
//   2. non-0 - failed assertion
func (me *THelper) ExpectExists(label string,
	subject interface{}, expected bool) int {
	exist := false

	switch x := subject.(type) {
	case bool:
		if x {
			exist = true
		}
	default:
		if x != nil {
			exist = true
		}
	}

	if (expected && exist) || (!expected && !exist) {
		return 0
	}

	s := label
	if s == "" {
		s = "object"
	}

	me.Errorf("%s - expect exist=%v, got=%v", s, expected, exist)

	return 1
}

// ExpectError asserts an error object existence.
//
// It takes 2 inputs:
//   1. subject     - the error test subject
//   2. expected    - the expectation for the subject existence.
//
// This function automatically logs the error message into if the test subject
// contradict the expectation.
//
// It returns:
//   1. 0     - passed assertion
//   2. non-0 - failed assertion
func (me *THelper) ExpectError(subject error, expected bool) int {
	return me.ExpectExists("error", subject, expected)
}

// SnapTime is a function to snap the current time.
//
// It returns:
//   1. Time        - time.Now() object.
func (me *THelper) SnapTime() *time.Time {
	s := time.Now()
	return &s
}

// CalculateDuration is a function to calculate time duration from a given
// start and a stop time.
//
// It calculates by taking the stop time subtracting the start time. Hence, if
// the 2 inputs are misplaced, the return value shall be negative.
//
// It takes 2 inputs:
//   1. start    - start time
//   2. stop     - stop time
//
// It returns:
//   1. duration - time duration
func (me *THelper) CalculateDuration(start *time.Time,
	stop *time.Time) time.Duration {
	if start == nil || stop == nil {
		return 0
	}

	return stop.Sub(*start)
}

// CalculateTimeLimits calculates the min and max time range for stop time.
//
// It first calculates the expected stop time which is by adding the expected
// duration time (`duration`) to the given start time (start).
//
// Then, It calculates:
//   1. the minimum limit by subtracting the stop time from the limit range
//      (`tolerance`).
//   2. the maximum limit by adding the limit range (range) into the stop time.
//
// Providing a `nil` to the start time will cause the function to do nothing.
// returning `nil` for both minimum and maximum values.
//
// It takes 2 inputs:
//   1. start       - the reference time
//   2. duration    - the expected duration for the stop time in nanoseconds
//   3. tolerance   - the symmetric time range between the stop time in
//                    nanoseconds.
//
// It returns:
//   1. minimum     - the timestamp at the stop time's minimum limit
//   2. maximum     - the timestamp at the stop time's maximum limit
func (me *THelper) CalculateTimeLimits(start *time.Time,
	duration int64,
	tolerance int64) (minimum *time.Time, maximum *time.Time) {
	if start == nil {
		return nil, nil
	}

	stop := start.Add(time.Duration(duration))
	l := stop.Add(time.Duration(-tolerance))
	t := stop.Add(time.Duration(tolerance))

	return &l, &t
}

// ExpectInTime asserts a given time is within a given set of time limits.
//
// It takes 1 main input and 2 optional inputs. They are:
//   1. `subject`  - the given test subject
//   2. `min`      - the minimum time limit (optional)
//   3. `max`      - the maximum time limit (optional)
//
// Depending on the availability of minimum and maximum, `ExpectInTime` carries
// out the assertion differently:
//   1. If both `min` and `max` are unavailable, it always return true since the
//      test subject is always in range.
//   2. If only `max` is available, it will check the test subject is always
//      below tha `max` limit. It will assert failed if it went over.
//   3. If only `min` is available, it will check the test subject is always
//      above the `min` limit. It will assert failed if it went under.
//   4. If both `max` and `min` are available, it will check the test subject
//      falls within range. It will assert failed if it went outside.
//   5. If test subject is not available, it will assert an `error`.
//
// This function automatically logs the error message into if the test subject
// contradict the expectation.
//
// It returns:
//   1. 0      - assertion is a passed case
//   2. not 0  - assertion is a failed case
func (me *THelper) ExpectInTime(subject, min, max *time.Time) int {
	switch {
	case subject == nil:
		me.Errorf("bad assertion - missing time subject")
	case max == nil && min == nil:
		return 0
	case max != nil && min == nil:
		return me.expectWithinMaxTimeLimit(subject, max)
	case max == nil && min != nil:
		return me.expectWithinMinTimeLimit(subject, min)
	case max != nil && min != nil:
		return me.expectWithinTimeRange(subject, min, max)
	}

	return 1
}

func (me *THelper) expectWithinTimeRange(subject, min, max *time.Time) int {
	if max.Sub(*subject) > 0 && subject.Sub(*min) > 0 {
		return 0
	}

	me.Errorf("out of range: %v | %v | %v", *min, *subject, *max)

	return 1
}

func (me *THelper) expectWithinMinTimeLimit(subject, min *time.Time) int {
	if subject.Sub(*min) > 0 {
		return 0
	}

	me.Errorf("over below limit: %v below %v", *subject, *min)

	return 1
}

func (me *THelper) expectWithinMaxTimeLimit(subject, max *time.Time) int {
	if max.Sub(*subject) > 0 {
		return 0
	}

	me.Errorf("over maximum limit: %v over %v", *subject, *max)

	return 1
}

// ExpectSameBytesSlices asserts 2 given bytes slices are the same.
//
// It checks for nil, length, and lastly byte by byte before returning a
// asserting its verdict.
//
// If both `a` and `b` are `nil`, the function shall pass the verdict since
// `nil` is the same as `nil`.
//
// It takes 4 inputs:
//   1. `labelA`    - naming for first slice (A) reporting purpose
//   2. `a`         - the first byte slice (A)
//   3. `labelB`    - naming for second slice (B) reporting purpose
//   4. `b`         - the second byte slice (B)
//
// This function automatically logs the error message into if the test subject
// contradict the expectation.
//
// It returns:
//   1. 0         - passed assertion
//   2. 1         - `a` is `nil` but `b` is **not** `nil`
//   3. 2         - `a` is **not** `nil` but `b` is `nil`
//   4. 3         - both `a` and `b` has different length
//   4. 4         - incorrect data matching during byte-to-byte comparison
func (me *THelper) ExpectSameBytesSlices(labelA string,
	a *[]byte,
	labelB string,
	b *[]byte) int {
	switch {
	case a == nil && b == nil:
		return 0
	case a == nil && b != nil:
		me.Errorf("%s is nil while %s is not nil",
			labelA,
			labelB,
		)

		return 1
	case a != nil && b == nil:
		me.Errorf("%s is nil while %s is not nil",
			labelB,
			labelA,
		)

		return 2
	}

	la := len(*a)
	lb := len(*b)

	if la != lb {
		me.Errorf("%s and %s has incorrect length: %v|%v",
			labelA,
			labelB,
			la,
			lb,
		)

		return 3
	}

	for i := 0; i < la; i++ {
		if (*a)[i] != (*b)[i] {
			me.Errorf("%s and %s has incorrect data: %v|%v",
				labelA,
				labelB,
				*a,
				*b,
			)

			return 4
		}
	}

	return 0
}

// ExpectUIDCorrectness asserts test UID is matching current index number.
//
// It takes 3 inputs:
//   1. `index`         - the current index number from the `for` loop.
//   2. `uid`           - the `UID` of the test case from its `Scenario`.
//   3. `startFromZero` - states whether the `UID` starts from 0.
//
// If the `startFromZero` is `true`, the `index` shall be incremented by 1
// internally.
//
// This function automatically logs the error message into if the test subject
// contradict the expectation.
//
// It returns:
//   1. 0             - successful assertion
//   2. non-zero      - failed assertion
func (me *THelper) ExpectUIDCorrectness(index int,
	uid int,
	startFromZero bool) (verdict int) {
	i := index

	if !startFromZero {
		i++
	}

	if i == uid {
		return 0
	}

	me.Errorf("uid mistmatched: index=%v uid=%v", i, uid)

	return 1
}

// ExpectSameStringSlices asserts 2 given string slices' contents are the same.
//
// It checks for `nil`, length, and lastly line by line before returning a
// its verdict.
//
// If both `a` and `b` are `nil`, the function shall pass the verdict since
// `nil` is the same as `nil`.
//
// It takes 4 inputs:
//   1. `labelA`    - name for first slice (A) reporting purpose
//   2. `a`         - the first string slice (A)
//   3. `labelB`    - name for second slice (B) reporting purpose
//   4. `b`         - the second string slice (B)
//
// This function automatically logs the error message into if the test subject
// contradict the expectation.
//
// It returns:
//   1. 0         - passed assertion
//   2. 1         - `a` is `nil` but `b` is **not** `nil`,
//   3. 2         - `a` is **not** `nil` but `b` is `nil`.
//   4. 3         - both `a` and `b` has different length.
//   4. 4         - incorrect data matching during line-by-line comparison
func (me *THelper) ExpectSameStringSlices(labelA string,
	a *[]string,
	labelB string,
	b *[]string) int {
	switch {
	case a == nil && b == nil:
		return 0
	case a == nil && b != nil:
		me.Errorf("%s is nil while %s is not nil",
			labelA,
			labelB,
		)

		return 1
	case a != nil && b == nil:
		me.Errorf("%s is nil while %s is not nil",
			labelB,
			labelA,
		)

		return 2
	}

	la := len(*a)
	lb := len(*b)

	if la != lb {
		me.Errorf("%s and %s has incorrect length: %v|%v",
			labelA,
			labelB,
			la,
			lb,
		)

		return 3
	}

	for i := 0; i < la; i++ {
		if (*a)[i] != (*b)[i] {
			me.Errorf("%s and %s has incorrect string line: %v|%v",
				labelA,
				labelB,
				*a,
				*b,
			)

			return 4
		}
	}

	return 0
}

// ExpectSameStrings asserts 2 given `string`s are having the same contents.
//
// This function automatically logs the error message into if the test subject
// contradict the expectation.
//
// It takes 4 inputs:
//   1. `labelA`    - name for first (a) string
//   2. `a`         - the first string
//   3. `labelB`    - name for second (b) string
//   4. `b`         - the second string
//
// It returns:
//   1. 0     - passed assertion
//   2. 1     - failed assertion
func (me *THelper) ExpectSameStrings(labelA string, subjectA string,
	labelB string, subjectB string) int {
	if subjectA != subjectB {
		me.Errorf("%s is not the same string as %s: %s|%s",
			labelA,
			labelB,
			subjectA,
			subjectB)

		return 1
	}

	return 0
}

// ExpectStringHasKeywords asserts a `string` contains a sub-string.
//
// This function automatically logs the error message into if the test subject
// contradict the expectation.
//
// It takes 4 inputs:
//   1. `labelA`         - label for the subject
//   2. `subject`        - the subject string
//   3. `labelB`         - label for the substring
//   4. `substring`      - the sub-string
//
// It returns:
//   1. 0              - passed assertion
//   2. 1              - failed assertion
func (me *THelper) ExpectStringHasKeywords(labelA string,
	subject string,
	labelB string,
	substring string) int {
	if strings.Contains(subject, substring) {
		return 0
	}

	me.Errorf("%s does not have substring %s: %v|%v",
		labelA,
		labelB,
		subject,
		substring)

	return 1
}

// ExpectSameBool asserts 2 given `bool`s are having the same contents.
//
// This function automatically logs the error message into if the test subject
// contradict the expectation.
//
// It takes 4 inputs:
//   1. `labelA`         - name for first (a) bool
//   2. `a`              - the first bool
//   3. `labelB`         - name for second (b) bool
//   4. `b`              - the second bool
//
// It returns:
//   1. 0              - passed assertion
//   2. 1              - failed assertion
func (me *THelper) ExpectSameBool(labelA string,
	subjectA bool,
	labelB string,
	subjectB bool) int {
	if subjectB == subjectA {
		return 0
	}

	me.Errorf("%s is not the same bool as %s: %v|%v",
		labelA,
		labelB,
		subjectA,
		subjectB)

	return 1
}

// ExpectSameFloat32 asserts 2 given `float32` are same.
//
// This function automatically logs the error message into if the test subject
// contradict the expectation.
//
// It takes 4 inputs:
//   1. `labelA`         - name for first (a) bool
//   2. `a`              - the first float32
//   3. `labelB`         - name for second (b) bool
//   4. `b`              - the second float32
//
// It returns:
//   1. 0              - passed assertion
//   2. 1              - failed assertion
func (me *THelper) ExpectSameFloat32(labelA string,
	subjectA float32,
	labelB string,
	subjectB float32) int {
	if subjectB == subjectA {
		return 0
	}

	me.Errorf("%s is not the same float32 as %s: %v|%v",
		labelA,
		labelB,
		subjectA,
		subjectB)

	return -1
}

// ExpectSameFloat64 asserts 2 given `float64` are same.
//
// This function automatically logs the error message into if the test subject
// contradict the expectation.
//
// It takes 4 inputs:
//   1. `labelA`         - name for first (a) bool
//   2. `a`              - the first `float64`
//   3. `labelB`         - name for second (b) bool
//   4. `b`              - the second `float64`
//
// It returns:
//   1. 0              - passed assertion
//   2. 1              - failed assertion
func (me *THelper) ExpectSameFloat64(labelA string,
	subjectA float64,
	labelB string,
	subjectB float64) int {
	if subjectB == subjectA {
		return 0
	}

	me.Errorf("%s is not the same float64 as %s: %v|%v",
		labelA,
		labelB,
		subjectA,
		subjectB)

	return -1
}

// ExpectSameBigFloat asserts 2 math/big.Float are same.
//
// This function automatically logs the error message into if the test subject
// contradict the expectation.
//
// It takes 4 inputs:
//   1. `labelA`         - name for first `*big.Float` (`a`)
//   2. `a`              - the first `*big.Float`
//   3. `labelB`         - name for second `*big.Float` (`b`)
//   4. `b`              - the second `*big.Float`
//
// It returns:
//   1. 0              - passed assertion (including `a` == `b` == `nil`)
//   2. 1              - failed assertion
func (me *THelper) ExpectSameBigFloat(labelA string,
	subjectA *big.Float,
	labelB string,
	subjectB *big.Float) int {
	switch {
	case subjectA == nil && subjectB != nil:
		me.Errorf("%s is nil while %s is not: %v|%v",
			labelA,
			labelB,
			subjectA,
			subjectB)

		return -1
	case subjectA != nil && subjectB == nil:
		me.Errorf("%s is nil while %s is not: %v|%v",
			labelB,
			labelA,
			subjectB,
			subjectA)

		return -1
	case subjectA == nil && subjectB == nil:
		return 0
	}

	if subjectA.Cmp(subjectB) == 0 {
		return 0
	}

	me.Errorf("%s is not the same *big.Float as %s: %v|%v",
		labelA,
		labelB,
		subjectA,
		subjectB)

	return -1
}
