// Copyright 2022 ZORALab Enterprise (hello@zoralab.com)
// Copyright 2022 "Holloway" Chew, Kean Ho (hollowaykeanho@gmail.com)
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package thelper

import (
	"fmt"
	"testing"
)

func Example() {
	t := &testing.T{}

	/***************************
	 * Basic Reporting         *
	 ***************************/
	// 1. Create Helper
	h := THelper{
		Controller:  t,
		FailKeyword: "TESTFAIL",
		QuietMode:   false,
	}

	// 2. Log messages
	h.Logf("This is a log message")
	h.Errorf("This is an error message that will fail the test")

	// 3. Conclude the test and print test messages at the end of the test
	h.Conclude()

	/****************************
	 * Assertion                *
	 ****************************/
	// 1. data existence
	a := []byte("Hello World")
	h.ExpectExists("demo bytes", a, true)  // pass
	h.ExpectExists("demo nil", nil, true)  // fail
	h.ExpectExists("demo nil", nil, false) // pass

	// 2. error existence
	err := fmt.Errorf("demo error")
	h.ExpectError(err, true)  // pass
	h.ExpectError(nil, true)  // fail
	h.ExpectError(nil, false) // pass

	// 3. Time sensitive assertion
	start := h.SnapTime()
	fmt.Printf("%v jumps over %v and made it to the %v",
		"James",
		"Lina",
		"Mountain")
	stop := h.SnapTime()
	myDuration := int64(300000) // 300 miliseconds
	minimum, maximum := h.CalculateTimeLimits(start, myDuration, 25000)
	h.ExpectInTime(stop, minimum, maximum)

	// 4. Byte Slices compare
	bSliceA := []byte{0xDE, 0xAD, 0xBE, 0xEF}
	bSliceB := []byte{0xde, 0xad, 0xbe, 0xef}
	h.ExpectSameBytesSlices("target", &bSliceA, "reference", &bSliceB)

	// 5. String Slices compare
	sSliceA := []string{"Hello", "world"}
	sSliceB := []string{"Bonjour", "World"}
	h.ExpectSameStringSlices("target", &sSliceA, "reference", &sSliceB)

	// 6. Same Strings compare
	h.ExpectSameStrings("subject", "Hello", "reference", "Bonjour")

	// 7. String Has Substring
	h.ExpectStringHasKeywords("subject", "Hello World", "keyword", "World")

	// 8. Same Boolean
	h.ExpectSameBool("subject", true, "reference", true)

	// 9. Capturing panic for sensitive executions
	var panicRet interface{}
	t.Run("s.stringUID()", func(t *testing.T) {
		defer func() {
			if r := recover(); r != nil {
				panicRet = r
			}
		}()

		h.ExpectSameStrings("subject", "Hello", "reference", "Bonjour")
	})
	h.Logf("%v\n", panicRet)
}
